"""
Appconfig for Django Clear To Build
"""

from django.apps import AppConfig


class DjangoCtbConfig(AppConfig):
    """
    Appconfig for Django Clear To Build
    """

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_ctb"
    verbose_name = "Clear To Build"



"""
Exceptions bruh
"""

from django_ctb.models import ProjectBuildPartShortage


class InsufficientInventory(Exception):
    """
    When there is not enough stock to fulfill a project build
    """

    def __init__(self, *args, shortages: list[ProjectBuildPartShortage], **kwargs):
        """
        Pass a list of shortages which were created to track this clear to
        build error.
        """
        self.shortages = shortages
        super().__init__(*args, **kwargs)


class MissingVendorPart(Exception):
    """
    When a vendor part cannot be found (maybe part number is wrong?)
    """


class RefNotFoundException(Exception):
    """
    The given commit ref could not be found in the git server
    """



"""
Services associated with placing and receiving orders from vendors.
"""

import logging
from dataclasses import dataclass

from django.utils import timezone

from django_ctb import models
from django_ctb.exceptions import MissingVendorPart

logger = logging.getLogger(__name__)


@dataclass
class _PartCount:
    part: models.Part
    count: int


class VendorOrderService:
    """
    Service for creating vendor orders, and seeing their fulfillment through
    to altering inventory stock. Provides traceability through inventory
    actions.
    """

    def _complete_order_line(self, order_line):
        # resolve part/inventory line
        inventory_line, _ = models.InventoryLine.objects.get_or_create(
            part=order_line.vendor_part.part, inventory=order_line.for_inventory
        )
        # create inventory action
        models.InventoryAction.objects.create(
            inventory_line=inventory_line,
            delta=order_line.quantity,
            order_line=order_line,
        )
        # update inventory line
        inventory_line.quantity += order_line.quantity
        inventory_line.save()

    def _complete_order(self, order):
        for order_line in order.lines.all():
            self._complete_order_line(order_line)
        order.fulfilled = timezone.now()
        order.save()

    def complete_order(self, order_pk):
        """
        Looks up a vendor order by PK then updates inventory lines for each
        part in the order, creates inventory action with change details.
        Marks vendor order as fulfilled.

        Ignores any vendor order marked fulfilled.
        """
        try:
            order = (
                models.VendorOrder.objects.filter(fulfilled__isnull=True)
                .prefetch_related("lines")
                .get(pk=order_pk)
            )
        except models.VendorOrder.DoesNotExist:
            raise
        self._complete_order(order)

    def _accumulate_shortfalls(self, build: models.ProjectBuild) -> list[_PartCount]:
        _shortfalls = {}
        for shortfall in build.shortfalls.all():
            part = shortfall.part
            _shortfalls.setdefault(part.pk, _PartCount(part=part, count=0))
            _shortfalls[part.pk].count += shortfall.quantity
        return list(_shortfalls.values())

    def _select_vendor_part(self, part: models.Part) -> models.VendorPart:
        logger.info(f">> Need more {part}, searching for best vendor")
        # TODO: should this incorporate the order volume somehow?
        vendor_part = (
            models.VendorPart.objects.filter(part=part).order_by("cost").first()
        )
        if vendor_part is None:
            logger.info(
                f"!! Part {part} does not have a vendor associated, "
                "cannot generate order!"
            )
            raise MissingVendorPart
        return vendor_part

    def _populate_vendor_order(
        self,
        *,
        vendor_part: models.VendorPart,
        quantity: int,
        inventory: models.Inventory,
    ):
        # get (or create) open vendor order for necessary vendor
        vendor_order, _ = models.VendorOrder.objects.get_or_create(
            vendor=vendor_part.vendor,
            placed__isnull=True,
        )
        # create order lines for shortfall
        order_line, _ = models.VendorOrderLine.objects.get_or_create(
            vendor_order=vendor_order,
            vendor_part=vendor_part,
            cost=vendor_part.cost,
            for_inventory=inventory,
            defaults={"quantity": 0},
        )
        order_line.quantity += quantity
        order_line.save()

    def generate_vendor_orders(self, build_pk):
        """
        Looks up a project build by PK then creates or updates vendor
        orders and order lines to cover shortfalls for the given project build.
        Searches for appropriate vendors for each part. Will silently ignore
        parts that don't have vendors.

        Ignores any project build which is completed.
        """
        # get shortfalls
        try:
            build = (
                models.ProjectBuild.objects.filter(completed__isnull=True)
                .prefetch_related("shortfalls")
                .get(pk=build_pk)
            )
        except models.ProjectBuild.DoesNotExist:
            return
        inventory = models.Inventory.objects.first()
        if inventory is None:
            return
        # analyze shortfalls for vendors and item numbers
        _shortfalls = self._accumulate_shortfalls(build)
        for _shortfall in _shortfalls:
            try:
                selected_vendor_part = self._select_vendor_part(_shortfall.part)
            except MissingVendorPart:
                # nothing to do for this part
                continue
            self._populate_vendor_order(
                vendor_part=selected_vendor_part,
                quantity=_shortfall.count,
                inventory=inventory,
            )



"""
Services and helpers for syncing bills of material (BOM) to project versions
"""

import csv
import io
import logging
from contextlib import closing

import requests
from django.db.models import Sum
from django.utils import timezone

from django_ctb import models
from django_ctb.exceptions import MissingVendorPart, RefNotFoundException
from django_ctb.github.services import GithubService
from django_ctb.mouser.services import MouserPartService

logger = logging.getLogger(__name__)


class ProjectVersionBomService:
    """Downloads BOM from repo at specified commit and creates project
    parts for each line."""

    def _get_vendor_part(self, *, row):
        try:
            vendor_part = models.VendorPart.objects.get(
                vendor__name=row.vendor_name, item_number=row.item_number
            )
        except models.VendorPart.DoesNotExist:
            # such a vendor part will need to exist before this project
            #  bom can be validated
            logger.info(
                f"Project BOM includes vendor specified part which is not "
                f"in parts library: {row.vendor_name}: {row.item_number}"
            )
            # If the vendor is Mouser then the part can be looked up via API
            if row.vendor_name == "Mouser":
                vendor_part = MouserPartService().create_vendor_part(
                    row=row,
                )
            else:
                raise MissingVendorPart(
                    f"Vendor Part not found {row.vendor_name}: {row.item_number}"
                )
        return vendor_part

    def _get_matching_parts(self, *, row):
        # Consider cases where there is more than one part that satisfies,
        #  e.g. an LED (parts may include a "white LED" and a "green LED").
        #  * Don't choose the part which is deprioritized
        #  * Choose the part which has inventory!
        return (
            models.Part.objects.filter(
                value=row.value,
                package__footprints__name=row.footprint_name,
                symbol__in=row.symbols,
            )
            .exclude(inventory_lines__is_deprioritized=True)
            .annotate(qty_in_inventory=Sum("inventory_lines__quantity"))
            .order_by("-qty_in_inventory")
        )

    def _get_part(self, *, row):
        if row.item_number and row.vendor_name:
            return self._get_vendor_part(row=row).part
        _part = self._get_matching_parts(row=row).first()
        return _part

    def _sync_footprints(serf, footprint_refs, *, project_part):
        # get rid of any outdated/altered refs
        models.ProjectPartFootprintRef.objects.filter(
            project_part=project_part
        ).exclude(footprint_ref__in=footprint_refs).delete()
        # create any missing refs
        for footprint_ref in footprint_refs:
            models.ProjectPartFootprintRef.objects.get_or_create(
                footprint_ref=footprint_ref, project_part=project_part
            )

    def _sync_implicit_parts(self, *, project_part):
        if project_part.part is None:
            logger.info(
                "!!!! Cannot create implicit parts for project_part lacking a "
                f"part {project_part}"
            )
            return
        implicit_project_parts = models.ImplicitProjectPart.objects.filter(
            for_package=project_part.part.package
        )
        project_part_pks = []
        for implicit_project_part in implicit_project_parts:
            quantity = implicit_project_part.quantity * project_part.quantity
            _implicit_project_part, _ = models.ProjectPart.objects.update_or_create(
                project_version=project_part.project_version,
                line_number=project_part.line_number,
                is_implicit=True,
                part=implicit_project_part.part,
                defaults={
                    "quantity": quantity,
                },
            )
            logger.info(
                f">>>> Created implicit project part {_implicit_project_part.pk}"
            )
            project_part_pks.append(_implicit_project_part.pk)
        # clean up any vestiges
        models.ProjectPart.objects.exclude(pk__in=project_part_pks).filter(
            line_number=project_part.line_number, is_implicit=True
        ).delete()

    def _sync_row(self, *, row, project_version):
        _part = None
        try:
            _part = self._get_part(row=row)
        except MissingVendorPart:
            pass
        defaults = {
            "part": _part,
            "quantity": row.quantity,
            "is_optional": row.optional,
        }
        if _part is None:
            defaults.update({"missing_part_description": f"{row}"})
        project_part, _ = models.ProjectPart.objects.update_or_create(
            project_version=project_version,
            line_number=row.line_number,
            is_implicit=False,
            defaults=defaults,
        )

        self._sync_footprints(row.references, project_part=project_part)
        self._sync_implicit_parts(project_part=project_part)
        return project_part

    def _get_commit_hash(self, project_version) -> str:
        ret = ""
        if project_version.project.git_server == models.Project.GitServer.GITHUB:
            ret = GithubService().get_commit_hash_for_ref(
                user=project_version.project.git_user,
                repo=project_version.project.git_repo,
                commit_ref=project_version.commit_ref,
            )
        return ret

    def _sync(self, *, project_version: models.ProjectVersion, synced_commit: str):
        row_errors = {}
        project_part_pks = []
        _bom_url = project_version.bom_url_for_commit(synced_commit)
        logger.info(f">> Getting BOM from {_bom_url}")
        file_response = requests.get(_bom_url)
        with (
            closing(file_response),
            io.StringIO(file_response.content.decode("utf-8")) as bom,
        ):
            reader = csv.DictReader(bom)
            logger.info(">> Parsing csv")
            for line_number, _row in enumerate(reader, start=1):
                if "#" not in _row:
                    _row["#"] = line_number
                row = models.BillOfMaterialsRow.model_validate(_row)
                logger.info(f">>>> Line {row.line_number}")
                project_part = self._sync_row(row=row, project_version=project_version)
                if project_part.part is None:
                    logger.info(f">>>> Part missing for line {row}")
                    logger.info(row.symbols)
                    row_errors.setdefault("part_missing", []).append(row.line_number)
                project_part_pks.append(project_part.pk)
        logger.info(f">> Created these project parts {project_part_pks}")
        # remove any outdated lines (implicit parts are cleaned up in
        #  _sync_implicit_parts)
        models.ProjectPart.objects.filter(
            project_version=project_version, is_implicit=False
        ).exclude(pk__in=project_part_pks).delete()
        project_version.last_synced_commit = synced_commit
        project_version.synced = timezone.now()
        project_version.save()
        return row_errors

    def sync(self, project_version_pk):
        """
        Finds a project version by PK then downloads BOM from repository and
        creates project parts for each row. BOM rows which cannot be matched
        to any part in the system will be flagged with a
        ``missing_part_description``. Parts available through supported
        vendors (Mouser) will be autopopulated provided the PartNum is known.

        Rows referencing footprints which are configured with implicit project
        parts will cause additional project parts to be created to represent
        the implicit parts.

        Upon completion of the sync process the commit hash where the BOM was
        found will be saved and the project version will be marked synced.
        """
        project_version = models.ProjectVersion.objects.get(pk=project_version_pk)
        logger.info(f"Starting project version sync for {project_version}")
        try:
            synced_commit = self._get_commit_hash(project_version=project_version)
        except RefNotFoundException:
            logger.info(
                f"!! Cannot find commit ref {project_version.commit_ref}. Aborting!"
            )
            return
        return self._sync(project_version=project_version, synced_commit=synced_commit)



"""
Services for interacting with the GitHub API
"""

import requests

from django_ctb.exceptions import RefNotFoundException


class GithubService:
    """
    Service for interacting with the GitHub API. Doesn't use a client, just
    calls directly for these minimal public endpoints.
    """

    base_url: str = "https://api.github.com"

    def _get_commit_hash(self, *, url_prefix: str, commit_ref: str) -> str:
        response = requests.get(f"{url_prefix}/commits/{commit_ref}")
        if response.status_code >= 300:
            # this is not a commit!
            raise RefNotFoundException
        # should be the same thing, but this seems right
        return response.json()["sha"]

    def _get_branch_head_commit_hash(self, *, url_prefix: str, commit_ref: str) -> str:
        response = requests.get(f"{url_prefix}/branches/{commit_ref}")
        if response.status_code >= 300:
            # this is not a branch!
            raise RefNotFoundException
        return response.json()["commit"]["sha"]

    def _get_tag_commit_hash(self, *, url_prefix: str, commit_ref: str) -> str:
        response = requests.get(f"{url_prefix}/tags")
        if response.status_code >= 300:
            # this is a real problem!
            raise RefNotFoundException
        for tag in response.json():
            if tag["name"] == commit_ref:
                return tag["commit"]["sha"]
        raise RefNotFoundException

    def get_commit_hash_for_ref(self, *, user: str, repo: str, commit_ref: str) -> str:
        """
        Find the commit has for the given ``commit_ref``. Uses the Github API to
        check, in this order, if ``commit_ref`` is:

        - A commit hash
        - A branch name
        - A tag

        The commit hash will be derived from the resource it represents.
        """
        # the `commit_ref` could be found at these paths
        # - a commit hash proper: `/repos/{owner}/{repo}/commits/{commit_sha}`
        # - a branch name: `/repos/{owner}/{repo}/branches/{branch_name}`
        # - a tag: `/repos/{owner}/{repo}/tags` <- Only list, no detail...
        url_prefix = f"{self.base_url}/repos/{user}/{repo}"
        try:
            return self._get_commit_hash(url_prefix=url_prefix, commit_ref=commit_ref)
        except RefNotFoundException:
            pass
        try:
            return self._get_branch_head_commit_hash(
                url_prefix=url_prefix, commit_ref=commit_ref
            )
        except RefNotFoundException:
            pass
        try:
            return self._get_tag_commit_hash(
                url_prefix=url_prefix, commit_ref=commit_ref
            )
        except RefNotFoundException:
            pass
        raise RefNotFoundException



"""
Dramatiq tasks for performing long-running actions in the background
"""

import dramatiq

from django_ctb.mouser.services import (
    populate_mouser_vendor_part,  # noqa: F401
)
from django_ctb.services import (
    ProjectBuildService,
    ProjectVersionBomService,
    VendorOrderService,
)


@dramatiq.actor
def sync_project_version(project_version_pk):
    """
    Background task to sync the Bill Of Materials (BOM) for a given project
    version. Syncing the BOM will create project parts for each BOM row which
    are linked to a part which matches the attributes of the row. When a
    suitable part cannot be found, a description of the part is included in the
    project part. Implicit project parts will be generated when a known
    footprint is called for (e.g. an LED footprint may create an implicit
    project part for the LED bezel).
    """
    ProjectVersionBomService().sync(project_version_pk)


@dramatiq.actor
def clear_to_build(project_build_pk):
    """
    Background task to reserve parts to cover a project build. Part
    reservations will remove stock from the inventory lines and create
    inventory actions to track. If there is insufficient stock to cover the
    project, a shortage will be associated with the build.
    """
    ProjectBuildService().clear_to_build(project_build_pk)


@dramatiq.actor
def complete_build(project_build_pk):
    """
    Background task to complete a cleared project build. All part reservations
    will be marked as utilized when the project build is completed.
    """
    ProjectBuildService().complete_build(project_build_pk)


@dramatiq.actor
def cancel_build(project_build_pk):
    """
    Background task to cancel a project build. All part reservations will be
    deleted, associated inventory lines credited, and inventory actions
    deleted.
    """
    ProjectBuildService().cancel_build(project_build_pk)


@dramatiq.actor
def generate_vendor_orders(project_build_pks):
    """
    Background task to create vendor orders from project build shortages.
    Prefers to add to existing (unplaced) orders rather than create a new one.
    """
    for project_build_pk in project_build_pks:
        VendorOrderService().generate_vendor_orders(project_build_pk)


@dramatiq.actor
def complete_order(vendor_order_pk):
    """
    Background task to complete a vendor order. All order lines will credit
    the appropriate inventory lines and inventory actions will be created to
    track.
    """
    VendorOrderService().complete_order(vendor_order_pk)



"""
Convenience collection of all main service classes
"""

from django_ctb.services.build import (
    PartSatisfactionManager,
    ProjectBuildPartReservationService,
    ProjectBuildService,
)
from django_ctb.services.order import (
    VendorOrderService,
)
from django_ctb.services.sync import (
    ProjectVersionBomService,
)

__all__ = [
    "InventoryLineFulfillment",
    "PartSatisfactionManager",
    "ProjectBuildPartReservationService",
    "ProjectBuildService",
    "ProjectVersionBomService",
    "VendorOrderService",
]



"""
Data models (and database models) for stock and projects.
"""

import re
from typing import TYPE_CHECKING

from django.db import models
from django.utils import timezone
from pydantic import AliasChoices, BaseModel, Field, field_validator

from django_ctb import conf  # noqa: F401

if TYPE_CHECKING:
    from django_stubs_ext.db.models.manager import RelatedManager


class Footprint(models.Model):
    """
    The manifestation of the part onto the printed circuit board. The
    footprint appears on the bill of materials, parts will be selected based
    in part on the footprint name match.
    """

    name = models.CharField(max_length=64)

    def __str__(self):  # pragma: no cover
        return self.name


class Package(models.Model):
    """
    The form factor for a part. E.g. Surface mount 0805, or TO-92.
    """

    class Technology(models.IntegerChoices):
        THROUGH_HOLE = 0
        SURFACE_MOUNT = 1
        UNKNOWN = 2

    technology = models.PositiveSmallIntegerField(
        choices=Technology, default=Technology.UNKNOWN
    )
    name = models.CharField(max_length=32, help_text="e.g. 0805, or TO-92W")
    footprints = models.ManyToManyField(Footprint, blank=True)

    def __str__(self):  # pragma: no cover
        return f"{self.get_technology_display()} {self.name}"  # type: ignore[unresolve-attribute]


class Vendor(models.Model):
    """
    Places where parts can be procured. e.g. Mouser, Tayda Electronics
    """

    name = models.CharField(max_length=64)
    base_url = models.URLField(help_text="sans trailing slash plz")

    def __str__(self):  # pragma: no cover
        return self.name


class Part(models.Model):
    """
    Individual parts which are available for procurement from a vendor and
    will be assembled into a project. e.g. a 100 Ohm surface mount (0805)
    resistor, or an NPN TO-92 transistor.

    The ``value`` attribute of the part will be matched to the ``value`` row
    on the bill of materials.
    """

    class Unit(models.IntegerChoices):
        NONE = 0
        OHM = 1
        FARAD = 2
        HENRY = 3
        VOLT = 4
        AMPERE = 5

    name = models.CharField(max_length=64)
    description = models.CharField(max_length=256, null=True, blank=True)
    value = models.CharField(
        max_length=32,
        help_text="Value as found on BOM e.g. '10K', '22u'",
        null=True,
        blank=True,
    )
    tolerance = models.SmallIntegerField(
        help_text="as percentage", null=True, blank=True
    )
    loading_limit = models.CharField(max_length=16, null=True, blank=True)
    unit = models.PositiveSmallIntegerField(choices=Unit, default=Unit.NONE)
    symbol = models.CharField(
        max_length=4,
        help_text="What prefix might this part have on a circuit schematic?",
        null=True,
        blank=True,
    )
    package = models.ForeignKey(
        Package, on_delete=models.PROTECT, null=True, blank=True
    )
    vendors = models.ManyToManyField(Vendor, related_name="parts", through="VendorPart")
    _is_scraped = models.BooleanField(default=False)
    equivalent_to = models.ForeignKey(
        "Part",
        on_delete=models.SET_NULL,
        related_name="equivalents",
        null=True,
        blank=True,
    )

    if TYPE_CHECKING:
        equivalents: RelatedManager["Part"]
        part_vendors: RelatedManager["VendorPart"]
        inventory_lines: RelatedManager["InventoryLine"]

    def __str__(self):  # pragma: no cover
        return f"{self.name} {self.symbol} {self.value} -- {self.package}"

    @property
    def unit_cost(self):
        """
        Extrapolated cost for each individual part
        """
        part_vendor = self.part_vendors.all().order_by("cost").first()
        if part_vendor is not None:
            return part_vendor.cost
        return 0


class ImplicitProjectPart(models.Model):
    """Certain parts do not appear on the BOM, but must be used for the final
    build. These are represented here. e.g. LED bezel, potentiometer knob, even
    the PCB could fit in this category?"""

    part = models.ForeignKey(Part, on_delete=models.PROTECT)
    for_package = models.ForeignKey(Package, on_delete=models.PROTECT)
    quantity = models.SmallIntegerField(default=1)


class VendorPart(models.Model):
    """
    The representation of a part as sold by a vendor. Pricing, item numbers,
    url paths are stored here.
    """

    vendor = models.ForeignKey(
        Vendor, related_name="vendor_parts", on_delete=models.PROTECT
    )
    part = models.ForeignKey(
        Part, related_name="part_vendors", on_delete=models.CASCADE
    )
    item_number = models.CharField(
        max_length=64, help_text="how the vendor identifies the part"
    )
    cost = models.DecimalField(
        decimal_places=4,
        max_digits=8,
        help_text="Cost per unit when purchased at given volume",
        null=True,
        blank=True,
    )
    volume = models.PositiveIntegerField(
        help_text="Must purchase this many to get cost",
        null=True,
        blank=True,
    )
    url_path = models.CharField(
        max_length=128, help_text="includes the leading slash!", null=True, blank=True
    )
    lot_cost = models.GeneratedField(
        expression=models.F("cost") * models.F("volume"),
        output_field=models.DecimalField(decimal_places=4, max_digits=8),
        db_persist=False,
    )

    @property
    def url(self):  # pragma: no cover
        """
        URL to the vendor part
        """
        return f"{self.vendor.base_url}{self.url_path}"

    def __str__(self):  # pragma: no cover
        return f"{self.vendor} - {self.item_number} - {self.part.name}"

    class Meta:
        ordering = ("vendor__name", "item_number")


class VendorOrder(models.Model):
    """
    Represents orders of parts from a vendor.
    """

    vendor = models.ForeignKey(Vendor, on_delete=models.PROTECT, related_name="orders")
    order_number = models.CharField(max_length=128, null=True, blank=True)
    created = models.DateTimeField(default=timezone.now)
    placed = models.DateTimeField(null=True, blank=True)
    fulfilled = models.DateTimeField(null=True, blank=True)

    def __str__(self):  # pragma: no cover
        return f"{self.vendor} #{self.order_number}"

    if TYPE_CHECKING:
        lines: RelatedManager["VendorOrderLine"]


class VendorOrderLine(models.Model):
    """
    Represents lines for individual parts in orders.
    """

    vendor_order = models.ForeignKey(
        VendorOrder, on_delete=models.CASCADE, related_name="lines"
    )
    vendor_part = models.ForeignKey(
        VendorPart, on_delete=models.PROTECT, related_name="lines"
    )
    quantity = models.PositiveIntegerField()
    cost = models.DecimalField(decimal_places=4, max_digits=8, help_text="per unit")
    for_inventory = models.ForeignKey("Inventory", on_delete=models.PROTECT)

    def __str__(self):  # pragma: no cover
        return (
            f"{self.vendor_order.vendor} {self.vendor_order.order_number} "
            f"x{self.quantity}"
        )


class Inventory(models.Model):
    """
    Container for complete stock of parts available for usage in project
    builds.
    """

    name = models.CharField(max_length=64)

    class Meta:
        verbose_name_plural = "Inventories"

    def __str__(self):  # pragma: no cover
        return self.name


class InventoryLine(models.Model):
    """
    Represents the stock of an individual part.
    """

    created = models.DateTimeField(default=timezone.now)
    updated = models.DateTimeField(auto_now=True)
    inventory = models.ForeignKey(Inventory, on_delete=models.PROTECT)
    part = models.ForeignKey(
        Part, on_delete=models.PROTECT, related_name="inventory_lines"
    )
    quantity = models.IntegerField(
        default=0,
        help_text="quantity on hand (unused reservations are removed from this number)",
    )
    is_deprioritized = models.BooleanField(default=False)

    if TYPE_CHECKING:
        inventory_actions: RelatedManager["InventoryAction"]

    @property
    def item_numbers(self) -> str:
        """
        Returns all item numbers from all vendors for the part in the
        inventory line (comma separated).
        """
        _item_numbers = []
        for vendor_part in self.part.part_vendors.all():
            _item_numbers.append(vendor_part.item_number)
        return ", ".join(_item_numbers)

    def __str__(self):  # pragma: no cover
        return f"{self.quantity}x {self.part} {self.item_numbers}"

    @property
    def quantity_on_hand(self) -> int:
        """
        Number of parts which are countable in physical inventory
        (includes numbers from pending or cleared reservations).
        """
        pending_quantity = sum(
            self.inventory_actions.filter(
                reservation__isnull=False, reservation__utilized__isnull=True
            ).values_list("delta", flat=True)
        )
        # values for delta are negative
        return self.quantity - pending_quantity


class InventoryAction(models.Model):
    """
    Tracks changes to inventory lines when orders are fulfilled and when
    project build parts are reserved.
    """

    inventory_line = models.ForeignKey(
        InventoryLine, on_delete=models.CASCADE, related_name="inventory_actions"
    )
    # what happened
    delta = models.IntegerField()
    # why did it happen (order fulfilled | project built | correction)
    order_line = models.ForeignKey(
        VendorOrderLine, on_delete=models.PROTECT, null=True, blank=True
    )
    # which reservation (unutilized | utilized)
    reservation = models.ForeignKey(
        "ProjectBuildPartReservation",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="inventory_actions",
    )
    # when did it happen
    created = models.DateTimeField(default=timezone.now)

    def __str__(self):  # pragma: no cover
        if self.order_line is not None:
            return (
                f"{self.delta} {self.inventory_line.part} from "
                f"{self.order_line.vendor_order.vendor.name}"
            )
        else:
            return f"{self.delta:+} {self.inventory_line.part}"


class Project(models.Model):
    """
    A thing you are building. This is a thin model with just a name and a url
    to a git repo. The repo must have a CSV file which is the Bill Of Materials
    (BOM) for the project. KiCAD generates such BOMs as a default feature.
    """

    class GitServer(models.IntegerChoices):
        UNKNOWN = 0
        GITHUB = 1

    name = models.CharField(max_length=64)

    git_server = models.PositiveSmallIntegerField(
        choices=GitServer, default=GitServer.GITHUB
    )
    git_user = models.CharField(max_length=64, null=True)
    git_repo = models.CharField(max_length=64, null=True)

    @property
    def git_url(self) -> str:
        """
        URL to the git repo represented by the project.
        """
        _url: str = "https://"
        if self.git_server == self.GitServer.GITHUB:
            _url += "github.com"
        _url += f"/{self.git_user}/{self.git_repo}"
        return _url

    def __str__(self):  # pragma: no cover
        return self.name


class ProjectVersion(models.Model):
    """
    A point-in-time representation of the project. Requires a commit ref
    (branch, tag, or commit hash) which exists in the repository, and the
    path within the repo to the bill of materials.
    """

    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    revision = models.IntegerField(default=0)
    commit_ref = models.CharField(
        max_length=64, help_text="Commit or tag representing the version"
    )
    bom_path = models.CharField(max_length=100)
    pcb_url = models.URLField(help_text="OSH park link to pcb", null=True, blank=True)
    pcb_cost = models.DecimalField(
        decimal_places=2, max_digits=6, help_text="for lot of 3", null=True, blank=True
    )
    synced = models.DateTimeField(null=True, blank=True)
    last_synced_commit = models.CharField(max_length=64, null=True, blank=True)

    def __str__(self):  # pragma: no cover
        return f"{self.project} v{self.revision}"

    if TYPE_CHECKING:
        project_parts: RelatedManager["ProjectPart"]

    @property
    def pcb_unit_cost(self) -> float:
        """
        The cost for a single PCB (assumes a PCB lot size of 3 per the minimum
        batch size at oshpark.com).
        """
        if self.pcb_cost is None:
            return 0.0
        return self.pcb_cost / 3

    @property
    def total_cost(self) -> float:
        """
        Extrapolates the full cost for all parts and PCB for this project
        version.
        """
        return self.pcb_unit_cost + sum([p.line_cost for p in self.project_parts.all()])

    @property
    def _bom_url_template(self) -> str:
        """
        The URL to the Bill of Materials computed from project and version data
        e.g. https://github.com/thismatters/EurorackLfo/raw/main/lfo.csv
        """
        _bom_path = self.bom_path.lstrip("/")
        return f"{self.project.git_url}/raw/{{commit_ref}}/{_bom_path}"

    @property
    def bom_url(self) -> str:
        """
        The URL to the Bill of Materials computed from project and version data
        e.g. https://github.com/thismatters/EurorackLfo/raw/main/lfo.csv
        """
        return self._bom_url_template.format(commit_ref=self.commit_ref)

    def bom_url_for_commit(self, commit_ref: str) -> str:
        """
        The URL to the Bill of Materials computed from project and version data
        using the given commit ref
        e.g. https://github.com/thismatters/EurorackLfo/raw/<commit_ref>/lfo.csv
        """
        return self._bom_url_template.format(commit_ref=commit_ref)


class ProjectPart(models.Model):
    """
    Representation of a BOM line for a project version. Holds references to the
    individual part, the footprint references (where the parts will be placed
    on the PCB), and the BOM line number. Allows for manual assignment of a
    substitute part.
    """

    part = models.ForeignKey(
        Part,
        on_delete=models.PROTECT,
        related_name="project_parts",
        null=True,
        blank=True,
    )
    substitute_part = models.ForeignKey(
        Part,
        on_delete=models.PROTECT,
        related_name="substitute_project_part",
        null=True,
        blank=True,
        help_text=(
            "When set, the substitute part will be used instead of the actual part"
        ),
    )
    missing_part_description = models.CharField(max_length=256, null=True, blank=True)
    project_version = models.ForeignKey(
        ProjectVersion, on_delete=models.CASCADE, related_name="project_parts"
    )
    line_number = models.SmallIntegerField()
    quantity = models.SmallIntegerField()
    is_implicit = models.BooleanField(default=False)
    is_optional = models.BooleanField(default=False)

    if TYPE_CHECKING:
        footprint_refs: RelatedManager["ProjectPartFootprintRef"]

    @property
    def line_cost(self):
        """
        Extrapolates the cost for the parts to satisfy this project part
        """
        if self.part is None:
            return 0
        if self.part.unit_cost is None:
            return 0
        return self.part.unit_cost * self.quantity

    @property
    def footprints(self):
        """
        Formats the footprint refrences into a comma-separated string.
        """
        _footprint_refs = self.footprint_refs.values_list("footprint_ref", flat=True)
        if self.is_optional:
            _footprint_refs = [_ref + "*" for _ref in _footprint_refs]
        return ", ".join(_footprint_refs)

    def __str__(self):
        _part = f"line {self.line_number} (part missing)"
        if self.part is not None:
            _part = str(self.part)
        return f"{_part} for {self.project_version}"


class ProjectPartFootprintRef(models.Model):
    """
    The actual, individual footprint ref where a project part will land on the
    PCB (e.g. R12)
    """

    project_part = models.ForeignKey(
        ProjectPart, on_delete=models.CASCADE, related_name="footprint_refs"
    )
    footprint_ref = models.CharField(max_length=8)

    def __str__(self):  # pragma: no cover
        return self.footprint_ref


class ProjectBuild(models.Model):
    """
    Represents a manufacturing run of a project version. Specify the number of
    instances of the project version that you will build.

    Any ``ProjectPart`` objects which were marked "optional" may be added to
    the ``excluded_project_parts``. When added, these project parts will not
    be omitted from clearing actvities.
    """

    project_version = models.ForeignKey(ProjectVersion, on_delete=models.PROTECT)
    quantity = models.SmallIntegerField()
    created = models.DateTimeField(default=timezone.now)
    cleared = models.DateTimeField(null=True, blank=True)
    completed = models.DateTimeField(null=True, blank=True)
    excluded_project_parts = models.ManyToManyField(
        ProjectPart,
        blank=True,
    )

    if TYPE_CHECKING:
        shortfalls: RelatedManager["ProjectBuildPartShortage"]

    @property
    def is_complete(self) -> bool:
        """
        Returns true if the project build has been completed.
        """
        return self.completed is not None

    def __str__(self):  # pragma: no cover
        suffix = ""
        if self.completed is not None:
            suffix = " [completed]"
        elif self.cleared is not None:
            suffix = " [cleared]"
        return f"{self.quantity}x {self.project_version}{suffix}"


class ProjectBuildPartShortage(models.Model):
    """
    Represents a part shortage which prevents a project build from being
    cleared.
    """

    part = models.ForeignKey(Part, on_delete=models.CASCADE)
    quantity = models.SmallIntegerField()
    project_build = models.ForeignKey(
        ProjectBuild, on_delete=models.CASCADE, related_name="shortfalls"
    )
    created = models.DateTimeField(default=timezone.now)


class ProjectBuildPartReservation(models.Model):
    """
    Reservations for parts to cover a project build.

    Reservations map to a ``Part`` and a ``ProjectPuild``.
    """

    # Okay, so. Project parts map to bom rows, and bom rows do not come with
    # any guarantees about part uniqueness across rows. So, it is possible for
    # more than one project part to share the same actual part. It is also
    # possible for there to be more than one inventory line associated with a
    # given project part (if, say, two equivalent parts have inventory lines
    # which only together cover the project, or if there are just two inventory
    # lines with the same exact part (for whatever reason)).

    project_build = models.ForeignKey(
        ProjectBuild, on_delete=models.CASCADE, related_name="part_reservations"
    )
    # We consolidate across the whole build on the ``part``, then create a
    # reservation which covers the full quantity (perhaps across several
    # inventory actions for equivalent lines as needed), and traceable to the
    # several project parts that share the same part.
    part = models.ForeignKey(
        Part,
        on_delete=models.SET_NULL,
        related_name="reservations",
        null=True,
    )

    # There may be more than one reservation per project part, given that
    # there is more than one build for a given project version.
    #
    # There may be more than one project part per reservation given more than
    # one bom row references the same part (e.g. when two jacks have a
    # different name).
    project_parts = models.ManyToManyField(
        ProjectPart, related_name="part_reservations"
    )
    # There may be more than one inventory action per reservation, given more
    # than one inventory line represents the part and any lacks sufficient
    # stock to complete the build alone (inventory actions cannot apply to
    # more than one reservation). (InventoryActions track the ``reservation``)

    created = models.DateTimeField(default=timezone.now)
    utilized = models.DateTimeField(null=True, blank=True)

    if TYPE_CHECKING:
        inventory_actions: RelatedManager[InventoryAction]

    @property
    def quantity(self) -> int:
        """
        The total quantity of parts encapsulated by this reservation (in a
        comma-separated string).
        """
        return sum(self.inventory_actions.all().values_list("delta", flat=True)) * -1

    @property
    def line_numbers(self) -> str:
        """
        All BOM line numbers encapsulated by this reservation (in a
        comma-sepaated string).
        """
        return ", ".join(
            [
                f"{line_num}"
                for line_num in self.project_parts.all().values_list(
                    "line_number", flat=True
                )
            ]
        )

    @property
    def footprints(self) -> str:
        """
        All footprint refs where parts will be placed (in a comma-separated
        string).
        """
        return (
            ", ".join(
                [
                    project_part.footprints
                    for project_part in self.project_parts.all()
                    if project_part.footprints
                ]
            )
            or "N/A"
        )


class BillOfMaterialsRow(BaseModel):
    """
    Maps to the default KiCAD BOM format with extra columns for "Vendor",
    "PartNum", and "Optional". Silently ignores any additional columns.
    """

    line_number: int | None = Field(validation_alias=AliasChoices("#", "line"))
    references: list[str] = Field(validation_alias=AliasChoices("Reference", "Ref"))
    quantity: int = Field(validation_alias=AliasChoices("Qty", "Qnty"))
    value: str = Field(alias="Value")
    footprint_name: str = Field(alias="Footprint")
    vendor_name: str | None = Field(alias="Vendor", default=None)
    item_number: str | None = Field(alias="PartNum", default=None)
    optional: bool = Field(alias="Optional", default=False)

    @property
    def symbols(self):
        """
        The footprint ref prefixes associated with the row. If the footprint
        refs for a line are "R1, R3", then the symbols will be "R".
        """
        return {r.strip("0123456789") for r in self.references}

    @field_validator("value", mode="before")
    @classmethod
    def normalize_value(cls, value: str):
        """Change "Value" column values to normalize potential formatting of
        component values with SI prefix __outside__ the value, and a period for
        a decimal separator:

        - 3M3 -> 3.3M (SI prefix greater than unity is capitalized)
        - 3U3 -> 3.3u (SI prefix less than unity is lowercased)
        - 1N4553 -> 1N4553 (Nano is excluded so that diodes don't get mangled)

        """
        matches = re.match(r"(?P<whole>\d+)(?P<prefix>[mMkKuUpPn])(?P<frac>\d+)", value)
        if matches is not None:
            _prefix = matches.group("prefix")
            if _prefix in "UP":
                _prefix = _prefix.lower()
            value = f"{matches.group('whole')}.{matches.group('frac')}{_prefix}"
        return value

    @field_validator("references", mode="before")
    @classmethod
    def split_and_strip(cls, value: str) -> list[str]:
        """
        Comma separated values in the "Reference" column will be split and
        stripped
        """
        return [v.strip() for v in value.split(",") if v.strip()]



"""
Services associated with Project Builds; clearing, and completing
"""

import logging

from django.db.models import QuerySet
from django.utils import timezone

from django_ctb import models
from django_ctb.exceptions import InsufficientInventory

logger = logging.getLogger(__name__)


class ProjectBuildPartReservationService:
    """
    Handles deletion of Project Build Part Reservations and restocking of
    inventory lines
    """

    def delete_reservation(self, reservation: models.ProjectBuildPartReservation):
        """
        Deletes reservation after crediting inventory lines and deleting
        inventory actions. Will not act on a utilized reservation.
        """
        if reservation.utilized is not None:
            return

        # undo inventory action
        for inventory_action in reservation.inventory_actions.all():
            inventory_action.inventory_line.quantity -= inventory_action.delta
            inventory_action.inventory_line.save()
            inventory_action.delete()

        reservation.delete()

    def delete_reservations(
        self,
        reservations: list[models.ProjectBuildPartReservation]
        | QuerySet[models.ProjectBuildPartReservation],
    ):
        """
        Deletes reservations after crediting inventory lines and deleting
        inventory actions. Will not act on a utilized reservation.
        """
        for reservation in reservations:
            self.delete_reservation(reservation)


class PartSatisfactionManager:
    """
    Manages satisfaction of combined demand for parts in a project build.
    This is a receptical in which to accumulate project parts which use a
    common part, then to assess the stock of said part (or it's equivalents),
    finally creating reservations or shortages.

    Use ``add_project_part`` to accumulate project parts, then
    use ``ensure_reservation`` to create a reservation (and inventory actions).
    """

    def __init__(
        self,
        *,
        part: models.Part,
        project_build: models.ProjectBuild,
    ):
        """
        Provide a part and a project build. Then use ``add_project_part`` to
        add any project parts which call for the same part. Once all project
        parts are tabulated, call ``ensure_reservation`` to create the actual
        reservation.
        """
        self.part = part
        self.project_build = project_build

        self.fulfilled: int = 0

        # These will be updated post initialization
        self.project_parts: list[models.ProjectPart] = []
        self.needed: int = 0

    @property
    def unfulfilled(self) -> int:
        """Number of parts required for the build which are not yet covered
        by reserved stock."""
        return self.needed - self.fulfilled

    def add_project_part(self, *, project_part: models.ProjectPart):
        """
        Add a project part for consideration in this manager. Increases the
        number of ``needed`` parts, and tracks the project part itself for
        association to the reservation (should it come to pass).
        """
        logger.info(
            "Adding project part: "
            f"{project_part.quantity} * {self.project_build.quantity}"
        )
        self.needed += project_part.quantity * self.project_build.quantity
        self.project_parts.append(project_part)

    def _ensure_inventory_action(
        self,
        *,
        inventory_line: models.InventoryLine,
        depletion: int,
        reservation: models.ProjectBuildPartReservation,
    ) -> models.InventoryAction:
        """
        Idempotent creation of inventory action and update of inventory line
        """
        inventory_action, _ = models.InventoryAction.objects.get_or_create(
            inventory_line=inventory_line,
            reservation=reservation,
            defaults={"delta": 0},
        )
        inventory_action.delta -= depletion
        inventory_action.save()
        inventory_line.quantity -= depletion
        inventory_line.save()
        return inventory_action

    @staticmethod
    def find_equivalent_parts(
        *,
        part: models.Part,
        depth: int = 0,
        maxdepth: int = 5,
    ) -> set[models.Part]:
        """
        Recursively searches for equivalent parts.
        """
        _equivalents = set()
        _equivalents.add(part)
        if depth >= maxdepth:
            return _equivalents
        if part.equivalent_to is not None:
            _equivalents.update(
                PartSatisfactionManager.find_equivalent_parts(
                    part=part.equivalent_to, depth=depth + 1, maxdepth=maxdepth
                )
            )
        others = part.equivalents.all()
        for other in others:
            _equivalents.update(
                PartSatisfactionManager.find_equivalent_parts(
                    part=other, depth=depth + 1, maxdepth=maxdepth
                )
            )
        return _equivalents

    def _get_inventory_lines_queryset(self):
        # determine equivalent parts
        _equivalents = PartSatisfactionManager.find_equivalent_parts(
            part=self.part,
        )

        # get inventory lines for all equivalent parts
        inventory_lines = models.InventoryLine.objects.filter(
            part__in=_equivalents, is_deprioritized=False
        ).order_by("quantity")
        return inventory_lines

    def _debit_inventory_for_reservation(
        self, reservation: models.ProjectBuildPartReservation
    ):
        """
        Handles situations where reservation quantities need to be increased.
        Will create/update inventory actions to cover the full reservation
        quantity.

        Takes stock such that there will be the fewest number of inventory
        lines with stock as possible.
        """
        logger.info(">> Debiting inventory")
        # deduct from inventory lines until need is fulfilled
        for inventory_line in self._get_inventory_lines_queryset():
            depletion = min(self.unfulfilled, inventory_line.quantity)
            self._ensure_inventory_action(
                inventory_line=inventory_line,
                depletion=depletion,
                reservation=reservation,
            )
            self.fulfilled += depletion
            logger.info(
                f">>>> Debiting inventory line {inventory_line} {depletion} parts"
            )
            if self.unfulfilled == 0:
                logger.info(">>>> Reservation satisfied")
                return

    def _credit_inventory_for_reservation(
        self,
        reservation: models.ProjectBuildPartReservation,
    ):
        """
        Return extra parts to stock from reservations such that there will be
        the fewest number of inventory lines with stock. Do not return stock
        to lines with zero stock (unless it is last resort)
        """
        logger.info(">> Crediting inventory")
        actions = reservation.inventory_actions.all().order_by(
            "-inventory_line__quantity", "delta"
        )
        for action in actions:
            # These should both be negative quantities
            credit = max(self.unfulfilled, action.delta) * -1
            action.delta += credit
            action.save()
            action.inventory_line.quantity += credit
            action.inventory_line.save()
            logger.info(
                f">>>> Crediting inventory line {action.inventory_line} {credit} parts"
            )
            if action.delta == 0:
                logger.info(">>>> Inventory action is depleted, deleting")
                action.delete()
            self.fulfilled -= credit
            if self.unfulfilled == 0:
                logger.info(">>>> Reservation satisfied")
                return

    def _check_stock(self):
        """
        Confirms that enough stock is on hand to cover the needed amount of
        parts. Raises ``InsufficientInventory`` otherwise.
        """
        logger.info(f">> Checking stock for {self.part}")
        inventory_lines = self._get_inventory_lines_queryset()
        total_stock = sum(inventory_lines.values_list("quantity", flat=True))
        if self.needed > total_stock:
            logger.info("!!!! Insufficient stock")
            self.fulfilled += total_stock
            shortage, _ = models.ProjectBuildPartShortage.objects.update_or_create(
                part=self.part,
                project_build=self.project_build,
                defaults={"quantity": self.unfulfilled},
            )
            raise InsufficientInventory(shortages=[shortage])

    def ensure_reservation(self) -> models.ProjectBuildPartReservation:
        """
        Idempotent creation of reservations for an individual part to satisfy
        build. Creates inventory actions to trace the reservation, and alters
        inventory line quantities to suit.
        """
        logger.info(f"Ensuring Reservation for {self.project_build} {self.part}")
        reservation = None
        try:
            reservation = models.ProjectBuildPartReservation.objects.get(
                project_build=self.project_build,
                part=self.part,
            )
            # add existing reservation quantity to fulfillment
            logger.info(
                f">> Reservation already exists with quantity {reservation.quantity}"
            )
            self.fulfilled += reservation.quantity
        except models.ProjectBuildPartReservation.DoesNotExist:
            pass

        self._check_stock()
        if reservation is None:
            logger.info(">> Creating Reservation")
            reservation = models.ProjectBuildPartReservation.objects.create(
                project_build=self.project_build,
                part=self.part,
            )
        reservation.project_parts.set(self.project_parts)

        if self.unfulfilled > 0:
            self._debit_inventory_for_reservation(reservation)
        elif self.unfulfilled < 0:
            self._credit_inventory_for_reservation(reservation)
        return reservation


class ProjectBuildService:
    """
    Service for clearing and completing project builds.
    """

    def _consolidate_project_parts(
        self, build: models.ProjectBuild
    ) -> list[PartSatisfactionManager]:
        """
        Consolidate project parts by the actual part (or substitute part)
        called for. Excludes excluded project parts.
        """
        # Get only parts which are actually included in the build
        excluded_project_part_pks = build.excluded_project_parts.all().values_list(
            "pk", flat=True
        )

        # Gather project parts by part (parts may be used on more than one row)
        # Accumulate total quantity required for all rows
        consolidated_project_parts: dict[int, PartSatisfactionManager] = {}
        for project_part in build.project_version.project_parts.all():
            if project_part.pk in excluded_project_part_pks:
                # Do not count this project part, it has been excluded from the build
                continue
            part = project_part.substitute_part or project_part.part
            consolidated_project_parts.setdefault(
                part.pk, PartSatisfactionManager(part=part, project_build=build)
            ).add_project_part(project_part=project_part)
        return list(consolidated_project_parts.values())

    def _clear_to_build(self, build) -> list[models.ProjectBuildPartReservation]:
        """
        Reserves sufficient stock of parts to complete a project, or---barring
        availability---reserves stock of parts which are pletiful enough to
        complete the project build and creates shortages for those unfortunate
        parts which have low stocks (then raises an ``InsufficientInventory``
        exception).
        """
        part_satisfactions = self._consolidate_project_parts(build)

        reservations: list[models.ProjectBuildPartReservation] = []
        shortages: list[models.ProjectBuildPartShortage] = []
        for part_satisfaction in part_satisfactions:
            try:
                reservations.append(part_satisfaction.ensure_reservation())
            except InsufficientInventory as exc:
                shortages.extend(exc.shortages)

        # clean up any resources left over from prior runs
        ProjectBuildPartReservationService().delete_reservations(
            models.ProjectBuildPartReservation.objects.exclude(
                id__in=[res.pk for res in reservations],
            ).filter(project_build=build)
        )
        models.ProjectBuildPartShortage.objects.exclude(
            id__in=[short.pk for short in shortages]
        ).filter(project_build=build).delete()

        # Display shortages for unavailable parts, and bail early
        if shortages:
            logger.info("!! Not clear to build !!")
            logger.info("!! Lacking: ")
            for shortage in shortages:
                _vendor_part = shortage.part.part_vendors.all().order_by("cost").first()
                logger.info(shortage.part, shortage.quantity)
                if _vendor_part:
                    logger.info(
                        f">> {_vendor_part.vendor.name} {_vendor_part.item_number}"
                    )
            raise InsufficientInventory(shortages=shortages)

        # Housekeeping
        build.cleared = timezone.now()
        build.save()
        return reservations

    def _complete_build(self, build):
        logger.info(f"Completing build {build}")
        if build.completed is not None:
            logger.info(f"!! Build already completed at {build.completed}")
            return
        reservations = self._clear_to_build(build)

        for reservation in reservations:
            reservation.utilized = timezone.now()
            reservation.save()

        build.completed = timezone.now()
        build.save()
        return

    def complete_build(self, build_pk):
        """
        Finds a project build by PK then marks it complete. This will mark any
        reservation associated with the project as utilized. Will only find
        project builds which are not complete and have been cleared.
        """
        try:
            build = (
                models.ProjectBuild.objects.filter(completed__isnull=True)
                .exclude(cleared__isnull=True)
                .select_related("project_version")
                .prefetch_related("project_version__project_parts")
                .get(pk=build_pk)
            )
        except models.ProjectBuild.DoesNotExist:
            raise
        return self._complete_build(build)

    def clear_to_build(self, build_pk):
        """
        Finds a project build by PK then reserves sufficient stock of parts to
        complete a project, or---barring availability---reserves stock of parts
        which are pletiful enough to complete the project build and creates
        shortages for those unfortunate parts which have low stocks.

        Ignores any project build which is completed.
        """
        try:
            build = (
                models.ProjectBuild.objects.filter(completed__isnull=True)
                .select_related("project_version")
                .prefetch_related("project_version__project_parts")
                .get(pk=build_pk)
            )
        except models.ProjectBuild.DoesNotExist:
            raise
        try:
            return self._clear_to_build(build)
        except InsufficientInventory:
            return []

    def _cancel_build(self, build):
        logger.info(f"Canceling build {build}")
        if build.completed is not None:
            logger.info("!! Build already completed, cannot cancel")
            return
        ProjectBuildPartReservationService().delete_reservations(
            build.part_reservations.all()
        )
        build.shortfalls.all().delete()
        build.cleared = None
        build.save()

    def cancel_build(self, build_pk):
        """
        Finds a project build by PK then removes any reservations or shortages
        associated with the project. Removes cleared status.

        Ignores any project build which is marked completed.
        """
        build = (
            models.ProjectBuild.objects.filter(completed__isnull=True)
            .prefetch_related("part_reservations")
            .get(pk=build_pk)
        )
        return self._cancel_build(build)



# Generated by Django 6.0.2 on 2026-02-22 13:39

import django.db.models.deletion
import django.db.models.expressions
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Footprint',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name='Inventory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
            ],
            options={
                'verbose_name_plural': 'Inventories',
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('git_server', models.PositiveSmallIntegerField(choices=[(0, 'Unknown'), (1, 'Github')], default=1)),
                ('git_user', models.CharField(max_length=64, null=True)),
                ('git_repo', models.CharField(max_length=64, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='ProjectBuild',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.SmallIntegerField()),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('cleared', models.DateTimeField(blank=True, null=True)),
                ('completed', models.DateTimeField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Vendor',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('base_url', models.URLField(help_text='sans trailing slash plz')),
            ],
        ),
        migrations.CreateModel(
            name='Package',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('technology', models.PositiveSmallIntegerField(choices=[(0, 'Through Hole'), (1, 'Surface Mount'), (2, 'Unknown')], default=2)),
                ('name', models.CharField(help_text='e.g. 0805, or TO-92W', max_length=32)),
                ('footprints', models.ManyToManyField(blank=True, to='django_ctb.footprint')),
            ],
        ),
        migrations.CreateModel(
            name='Part',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('description', models.CharField(blank=True, max_length=256, null=True)),
                ('value', models.CharField(blank=True, help_text="Value as found on BOM e.g. '10K', '22u'", max_length=32, null=True)),
                ('tolerance', models.SmallIntegerField(blank=True, help_text='as percentage', null=True)),
                ('loading_limit', models.CharField(blank=True, max_length=16, null=True)),
                ('unit', models.PositiveSmallIntegerField(choices=[(0, 'None'), (1, 'Ohm'), (2, 'Farad'), (3, 'Henry'), (4, 'Volt'), (5, 'Ampere')], default=0)),
                ('symbol', models.CharField(blank=True, help_text='What prefix might this part have on a circuit schematic?', max_length=4, null=True)),
                ('_is_scraped', models.BooleanField(default=False)),
                ('equivalent_to', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='equivalents', to='django_ctb.part')),
                ('package', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_ctb.package')),
            ],
        ),
        migrations.CreateModel(
            name='InventoryLine',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('quantity', models.IntegerField(default=0, help_text='quantity on hand (unused reservations are removed from this number)')),
                ('is_deprioritized', models.BooleanField(default=False)),
                ('inventory', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_ctb.inventory')),
                ('part', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='inventory_lines', to='django_ctb.part')),
            ],
        ),
        migrations.CreateModel(
            name='ImplicitProjectPart',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.SmallIntegerField(default=1)),
                ('for_package', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_ctb.package')),
                ('part', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_ctb.part')),
            ],
        ),
        migrations.CreateModel(
            name='ProjectBuildPartShortage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.SmallIntegerField()),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('part', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ctb.part')),
                ('project_build', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='shortfalls', to='django_ctb.projectbuild')),
            ],
        ),
        migrations.CreateModel(
            name='ProjectPart',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('missing_part_description', models.CharField(blank=True, max_length=256, null=True)),
                ('line_number', models.SmallIntegerField()),
                ('quantity', models.SmallIntegerField()),
                ('is_implicit', models.BooleanField(default=False)),
                ('is_optional', models.BooleanField(default=False)),
                ('part', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='project_parts', to='django_ctb.part')),
                ('substitute_part', models.ForeignKey(blank=True, help_text='When set, the substitute part will be used instead of the actual part', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='substitute_project_part', to='django_ctb.part')),
            ],
        ),
        migrations.CreateModel(
            name='ProjectBuildPartReservation',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('utilized', models.DateTimeField(blank=True, null=True)),
                ('part', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='reservations', to='django_ctb.part')),
                ('project_build', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='part_reservations', to='django_ctb.projectbuild')),
                ('project_parts', models.ManyToManyField(related_name='part_reservations', to='django_ctb.projectpart')),
            ],
        ),
        migrations.AddField(
            model_name='projectbuild',
            name='excluded_project_parts',
            field=models.ManyToManyField(blank=True, to='django_ctb.projectpart'),
        ),
        migrations.CreateModel(
            name='ProjectPartFootprintRef',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('footprint_ref', models.CharField(max_length=8)),
                ('project_part', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='footprint_refs', to='django_ctb.projectpart')),
            ],
        ),
        migrations.CreateModel(
            name='ProjectVersion',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('revision', models.IntegerField(default=0)),
                ('commit_ref', models.CharField(help_text='Commit or tag representing the version', max_length=64)),
                ('bom_path', models.CharField(max_length=100)),
                ('pcb_url', models.URLField(blank=True, help_text='OSH park link to pcb', null=True)),
                ('pcb_cost', models.DecimalField(blank=True, decimal_places=2, help_text='for lot of 3', max_digits=6, null=True)),
                ('synced', models.DateTimeField(blank=True, null=True)),
                ('last_synced_commit', models.CharField(blank=True, max_length=64, null=True)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ctb.project')),
            ],
        ),
        migrations.AddField(
            model_name='projectpart',
            name='project_version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='project_parts', to='django_ctb.projectversion'),
        ),
        migrations.AddField(
            model_name='projectbuild',
            name='project_version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_ctb.projectversion'),
        ),
        migrations.CreateModel(
            name='VendorOrder',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order_number', models.CharField(blank=True, max_length=128, null=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('placed', models.DateTimeField(blank=True, null=True)),
                ('fulfilled', models.DateTimeField(blank=True, null=True)),
                ('vendor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='orders', to='django_ctb.vendor')),
            ],
        ),
        migrations.CreateModel(
            name='VendorOrderLine',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveIntegerField()),
                ('cost', models.DecimalField(decimal_places=4, help_text='per unit', max_digits=8)),
                ('for_inventory', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_ctb.inventory')),
                ('vendor_order', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='lines', to='django_ctb.vendororder')),
            ],
        ),
        migrations.CreateModel(
            name='InventoryAction',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('delta', models.IntegerField()),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('inventory_line', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='inventory_actions', to='django_ctb.inventoryline')),
                ('reservation', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='inventory_actions', to='django_ctb.projectbuildpartreservation')),
                ('order_line', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_ctb.vendororderline')),
            ],
        ),
        migrations.CreateModel(
            name='VendorPart',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('item_number', models.CharField(help_text='how the vendor identifies the part', max_length=64)),
                ('cost', models.DecimalField(blank=True, decimal_places=4, help_text='Cost per unit when purchased at given volume', max_digits=8, null=True)),
                ('volume', models.PositiveIntegerField(blank=True, help_text='Must purchase this many to get cost', null=True)),
                ('url_path', models.CharField(blank=True, help_text='includes the leading slash!', max_length=128, null=True)),
                ('lot_cost', models.GeneratedField(db_persist=False, expression=django.db.models.expressions.CombinedExpression(models.F('cost'), '*', models.F('volume')), output_field=models.DecimalField(decimal_places=4, max_digits=8))),
                ('part', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='part_vendors', to='django_ctb.part')),
                ('vendor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='vendor_parts', to='django_ctb.vendor')),
            ],
            options={
                'ordering': ('vendor__name', 'item_number'),
            },
        ),
        migrations.AddField(
            model_name='vendororderline',
            name='vendor_part',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='lines', to='django_ctb.vendorpart'),
        ),
        migrations.AddField(
            model_name='part',
            name='vendors',
            field=models.ManyToManyField(related_name='parts', through='django_ctb.VendorPart', to='django_ctb.vendor'),
        ),
    ]



"""
Config for Django Clear To Build
"""

from appconf import AppConf
from django.conf import settings  # noqa: F401


class DjangoCTBAppConf(AppConf):
    """
    Config for Django Clear To Build
    """

    MOUSER_API_KEY = ""

    class Meta:
        prefix = "ctb"



"""
A package for tracking inventory for small-scale electronics manufacturing.
It tracks parts, vendors, orders, inventories (the parts that you have
on-hand), projects across several versions (and their bills of materials,
including cost breakdowns), and finally provides clearance to build---indicating
that your inventory of parts is sufficient to complete the project build.
"""

from importlib.metadata import PackageNotFoundError, version

__title__ = "Django Clear To Build"
__author__ = "Paul Stiverson"
__license__ = "GNU Affero General Public License v3"
__copyright__ = "Copyright 2024 Paul Stiverson"

try:
    __version__ = version("django_ctb")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0"
